
#include <dlfcn.h>

#include "Tracer.h"

#include <vector>
#include <strstream.h>

#include <errstr.h>
#include <types.h>


#include "scanner.h"
#include "dal_utilities.h"
#include "PerlInterpreter.h"
#include "parsing.h"
#include "param.h"
#include "dal_c_lookups.h"
#include "SasFileLocator.h"
#include "deceit_lookup_table.h"
#include "perl_deceit_lookup_table.h"

#include "deceit.h" // generated automatically by the make
#include "perlutils.h"

// task main body at the end

template<class T> static void
setElement(
		Column *	column,
		unsigned int	elementIndex,
		T		value )
{
	PROC

	ColumnData * data = column -> data();

	switch( column -> dataType() )
	{
		case Data::Int8 :

			data -> int8Data()[ elementIndex ] = value;

			break;

		case Data::Int16 :

			data -> int16Data()[ elementIndex ] = value;

			break;

		case Data::Int32 :

			data -> int32Data()[ elementIndex ] = value;

			break;

		case Data::Real32 :

			data -> real32Data()[ elementIndex ] = value;

			break;

		case Data::Real64 :

			data -> real64Data()[ elementIndex ] = value;

			break;

		case Data::Bool :

			break;

		case Data::DString :

			break;

		default :

			errstr << Sas::fatal("program error");
			break;

	} // switch

	delete data;

	ENDPROC

}


static const string		dalBlock = "execute_dal";
static const string		perlBlock = "execute_perl";
static const string		postBlock = "execute_post";


void
fillColumn(
	Column *	column,
	const char *	fileName,
	int		columnNumber )
{
	PROC

	ifstream	is( fileName );

	if( !is )
	{
		strstream	errorMessage;

		errorMessage << "The file " << fileName

			<< " could not be opened" << ends;

		errstr << Sas::fatal(errorMessage.str());

	} // if

	unsigned int	elementIndex = 0;

	while( !is.eof() )
	{
		char	currentLine[ 1024 ];
		Token	token;

		is.getline( currentLine, 1000 );

		istrstream	lineToScan( currentLine, strlen( currentLine ));

		//currentLine[ strlen( currentLine ) ] = EOF;

		ScannerT	scanner( lineToScan );


		for( int i = 0; i < columnNumber; ++i ) token = scanner.acceptNext();

		token = scanner.acceptNext();

		switch( token.id() )
		{
			case Token::Integer :

				setElement(
					column,
					elementIndex,
					token.asInt() );

				break;

			case Token::EndOfFile :

				break;

			default :

				errstr<< Sas::fatal("bad data type");

				break;

		} // switch

		elementIndex++;

	} // while
	
	ENDPROC

}


void
fillTable(
	Table *		table,
	const char *	fileName )
{
	PROC

	const unsigned int NrOfColumns = table -> columns();

	for(
		unsigned int i = 0;
		i < NrOfColumns;
		++i )
	{
		fillColumn(
			table -> column( i ),
			fileName,
			i );

	} // for

	ENDPROC

}


static struct Lookups lookups[] =
{
{ string( "addDLFunctionToLookupTable"),	Token( Functor( Functor::C, (void *)addDLFunctionToLookupTable, Token::Undefined ))},
{ string( "addPerlFunctionToLookupTable"),	Token( Functor( Functor::C, (void *)addPerlFunctionToLookupTable,		Token::Undefined ))},
{ string( "fillColumn" ),	Token( Functor( Functor::Perl, (void *)fillColumn, 	int(0) ))},
{ string( "fillTable" ),	Token( Functor( Functor::C, (void *)fillTable,		int(0) ))},
{ string( "rows" ),		Token( Functor( Functor::Perl, 0,			int(0) ))},
};


static TableT
constructLookupTable()
{
	PROC

	TableT	lookupTable;

	perlInterpreter.addFunctionsToTable( lookupTable );
	lookupTable.append( dalCLookups() );

	const unsigned int numberOfFunctions =

				sizeof lookups /
				sizeof( Lookups );

	for( unsigned int i = 0; i < numberOfFunctions; ++i )
	{
		lookupTable.Insert( lookups[ i ] .name, lookups[ i ] .token );
	} // for

	RETURN lookupTable;

}

static TableT			LookupTable = constructLookupTable();
						

void
addPerlFunctionToLookupTable( const char * functionName )
{
	PROC
	ostrstream exp;
	exp << functionName << "=perlFunctor()" << ends;
	eval( exp.str(), &LookupTable );
	ENDPROC
}


void
//addDLFunctionToLookupTable( const vector<Token>& arguments )
addDLFunctionToLookupTable(
		const char * 	libraryName,
		const char * 	functionName,
		DataType	returnType )
{
	PROC
	static map<string,void*> handleOf;

	//const string	libraryName = arguments[ 0 ].asString();
	//const string    functionName = arguments[ 1 ].asString();
	//const DataType	returnType = (DataType)( arguments[ 2 ].asInt() );
	//cout << "returnType = " << int(returnType) << endl;

	map<string,void*>::iterator h = handleOf.find( libraryName );
	if( h == handleOf.end() )
	{
		handleOf[ libraryName ] = dlopen( libraryName, RTLD_LAZY );
	} // if

	void * libraryHandle = handleOf[ libraryName ];

	ostrstream exp;
	void (*fptr)() = (void (*)())(dlsym(libraryHandle, functionName));
	//double (*Cosine)(double) = (double(*)(double))(dlsym(libraryHandle, functionName.c_str()));

	//cout << "cos(0.5) = " << Cosine(0.5) << endl;

	exp << functionName << "=CFunctor(" << (int)(fptr) << "," << int(returnType) << ")" << ends;
	//cout << "exp = " << exp.str() << endl;

	eval( exp.str(), &LookupTable );

}


static void
expectedOneOf( ScannerT & scanner, const char * string1, ... )
{
	PROC
	va_list		args;
	char *		arg;
	string		strings( string1 );

	va_start( args, string1 );

	while( (arg = va_arg( args, char * )) != 0 )
	{
		strings += string( ", " ) + arg;
	} // while

	va_end( args );

	StreamT & file = scanner.stream();

	errstr << file.FileNameOf() << ": Error at line " <<
		file.CurrentLineNo() <<
 		", and at or near column " <<
		file.CurrentColumnNo() <<
		" - Expected one of " << strings << Sas::fatal("expected one of");

	ENDPROC
}


static void
parseExecuteDal( ScannerT & scanner )
{
	PROC

	Token			token;
	Token			result;
	Token::BTree		expression;

	scanner.match( dalBlock );
	scanner.match( '<' );

	while( !scanner.NextIs( '>' ))
	{
		expression = ParseStatement( scanner );
		perlInterpreter.EvaluateExpression( expression, LookupTable );

	} // while

	scanner.match( '>' );

	ENDPROC

}


static void
parsePostBlock( ScannerT & scanner, vector<Token::BTree> & expressions )
{
	PROC

	Token			token;
	Token			result;
	Token::BTree		expression;

	scanner.match( postBlock );
	scanner.match( '<' );

	while( !scanner.NextIs( '>' ))
	{
		expression = ParseStatement( scanner );
		expressions.push_back( expression );

	} // while

	scanner.match( '>' );

	ENDPROC

}


static void
parseExecutePerl( ScannerT & scanner )
{
	PROC

	Token			token;
	Token			statements;

	scanner.match( perlBlock );
	scanner.SetState( ScannerT::FREE_FORMAT_TEXT );

	scanner.match( '<' );

	scanner.SetState( ScannerT::NORMAL );
	scanner.Match( Token::FreeText, statements );
	scanner.match( '>' );
	evalperl( statements.asString().c_str() );

	ENDPROC

}


static void
parseText(
		ScannerT &	scanner,
		Token &		text )
{
	PROC
	// ***********************************************************
	// The text may be a single line literal string, delimited
	// by the usual quotation marks '"', or may be a multiline
	// free format piece of text, delimited by '<' and '>'.
	// ***********************************************************
	if( scanner.NextIs( '<' ) )
	{
		scanner.SetState( ScannerT::FREE_FORMAT_TEXT );
		scanner.match( '<' );
		scanner.SetState( ScannerT::NORMAL );
		scanner.Match( Token::FreeText, text );
		scanner.match( '>' );
	}
	else
	{
		scanner.Match( Token::String, text );
	} // if

	ENDPROC
}


static void
parseCommentDescriptor(
		ScannerT &	scanner,
		Attributable *	attributable )
{
	PROC

	Token		text;

	scanner.match( "comment" );
	parseText( scanner, text );
	attributable -> addComment( text.asString() );

	ENDPROC

}


static void
parseHistoryDescriptor(
		ScannerT &	scanner,
		Attributable *	attributable )
{
	PROC

	Token		text;

	scanner.match( "history" );
	parseText( scanner, text );
	attributable -> addHistory( text.asString() );

	ENDPROC

}


static void
parseAttributeDescriptorBody(
		ScannerT &		scanner,
		Attribute *		attribute,
		vector<Token::BTree> &	postExpressions )
{
	PROC

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.nextIs( perlBlock ))
		{
			parseExecutePerl( scanner );
		}
		else if( scanner.nextIs( dalBlock ))
		{
			parseExecuteDal( scanner );
		}
		else if( scanner.nextIs( postBlock ))
		{
			parsePostBlock( scanner, postExpressions );
		}
		else if( scanner.NextIs( "include" ))
		{
			Token	fileName;

			// ************************************************************
			// The include directive must appear on a single line.
			// Get the line, and parse it.
			// ************************************************************
			scanner.SetState( ScannerT::LINE );
			scanner.match( "include" );
			scanner.Match( Token::String, fileName );
			scanner.SetState( ScannerT::NORMAL );

			// **********************************************************
			// Initialise a scanner and parse and execute the specified
			// include file.
			// **********************************************************
			ScannerT	s( fileName.asString() );
			parseAttributeDescriptorBody( s, attribute, postExpressions );
		}
		else
		{
			expectedOneOf(
				scanner,
				"include",
				perlBlock.c_str(),
				dalBlock.c_str(),
				postBlock.c_str(),
				0 );
		} // if
	} // while

	ENDPROC
}


static void
parseAttributeDescriptor(
		ScannerT &		scanner,
		Attributable *		attributeSet )
{
	PROC

	Token			token;
	Token			name;
	Token			type;
	Token			value;
	Token			comment;
	Token			units;
	Token::BTree		expression;
	bool			expressionDetected = false;
	vector<Token::BTree>	postExpressions;

	// ************************************************************
	// The attribute descriptor must appear on a single line.
	// Get the line, and parse it.
	// ************************************************************
	scanner.SetState( ScannerT::LINE );
	scanner.match( "attribute" );
	expression = ParseExpression( scanner );
	name = perlInterpreter.EvaluateExpression( expression, LookupTable );
	type = scanner.MatchOneOf(
			"int",
			"real",
			"string",
			"bool",
			(char *)0 );
	expression = ParseExpression( scanner );
	value = perlInterpreter.EvaluateExpression( expression, LookupTable );
	scanner.Match( Token::String, comment );
	scanner.Match( Token::String, units );

	// ***************************************
	// Add the attribute to the attribute set.
	// ***************************************
	Attribute * newAttribute =

		attributeSet -> addAttribute(
					name.asString(),
					comment.asString(),
					units.asString() );

	// ***********************************************************
	// Assign the new attribute to the lookup table variable "att"
	// "att" is really a pointer, but treat it as an integer
	// which should be ok (for this prototype version), but
	// later this should be implemented using a Pointer token.
	// ********************************************************
	ostrstream exp;
	exp << "att=pointer(" << int( newAttribute ) << ")" << ends;
	eval( exp.str(), &LookupTable );

	// **********************************
	// Check for the optional expression.
	// **********************************
	if( !scanner.EndOfFile() )
	{
		expression = ParseExpression( scanner );
		expressionDetected = true;
	} // if

	scanner.SetState( ScannerT::NORMAL );

	// ****************************
	// Now set the attribute value.
	// ****************************
	switch( toAttributeDataType( type.asString() ))
	{
		case Attribute::Int :

			*newAttribute = value.asInt();

			break;

		case Attribute::Real :

			*newAttribute = value.asReal();

			break;

		case Attribute::DString :

			*newAttribute = value.asString();

			break;

		case Attribute::Bool :

			*newAttribute = value.asBool();
			break;

		default :

			errstr << Sas::fatal("Bad Attribute Type");

			break;

	} // switch

	if( scanner.NextIs( '<' ))
	{
		scanner.match( '<' );

		parseAttributeDescriptorBody( scanner, newAttribute, postExpressions );

		scanner.match( '>' );
	} // if

	// ***********************************
	// Evaluate post expressions (if any).
	// ***********************************
	for( unsigned int i = 0; i < postExpressions.size(); ++i )
	{
		Token::BTree		expression;
		expression = postExpressions[ i ];
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // for

	if( expressionDetected )
	{
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // if

	ENDPROC

}


static void
parseColumnDescriptorBody(
		ScannerT &	scanner,
		Column *	column,
		vector<Token::BTree> &	postExpressions )
{
	PROC

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.NextIs( "attribute" ))
		{
			parseAttributeDescriptor(
				scanner,
				column );
		}
		else if( scanner.nextIs( perlBlock ))
		{
			parseExecutePerl( scanner );
		}
		else if( scanner.nextIs( dalBlock ))
		{
			parseExecuteDal( scanner );
		}
		else if( scanner.nextIs( postBlock ))
		{
			parsePostBlock( scanner, postExpressions );
		}
		else if( scanner.NextIs( "include" ))
		{
			Token	fileName;

			// ************************************************************
			// The include directive must appear on a single line.
			// Get the line, and parse it.
			// ************************************************************
			scanner.SetState( ScannerT::LINE );
			scanner.match( "include" );
			scanner.Match( Token::String, fileName );
			scanner.SetState( ScannerT::NORMAL );

			// **********************************************************
			// Initialise a scanner and parse and execute the specified
			// include file.
			// **********************************************************
			ScannerT	s( fileName.asString() );
			parseColumnDescriptorBody( s, column, postExpressions );
		}
		else
		{
			expectedOneOf(
				scanner,
				"attribute",
				"include",
				perlBlock.c_str(),
				dalBlock.c_str(),
				postBlock.c_str(),
				0 );
		} // if
	} // while

	ENDPROC
}


static void
parseColumnDescriptor(
		ScannerT &	scanner,
		Table *		table )
{

	PROC

	Token			token;
	Token			name;
	Token			type;
	Token			label;
	Token			units;
	Token			size;
	vector<unsigned long>	dimensionSizes;
	Column *		column;
	Token::BTree		expression;
	bool			expressionDetected = false;
	vector<Token::BTree>	postExpressions;

	// ************************************************************
	// The column descriptor must appear on a single line.
	// Get the line, and parse it.
	// ************************************************************
	scanner.SetState( ScannerT::LINE );
	scanner.match( "column" );

	// *****************************************************************
	// Check for optional dimension sizes.
	// If there are any, each dimension size is added to dimensionSizes.
	// If there are none (which means that the column is scalar),
	// then set dimensionSizes to Dal::scalar
	// *****************************************************************
	if( scanner.NextIs( '(' ) )
	{
		scanner.match( '(' );
		expression = ParseAssignmentExpression( scanner );
		size = perlInterpreter.EvaluateExpression( expression, LookupTable );
		dimensionSizes.push_back( size.asInt() );

		while( scanner.NextIs( ',' ) )
		{
			scanner.match( ',' );
			expression = ParseAssignmentExpression( scanner );
			size = perlInterpreter.EvaluateExpression( expression, LookupTable );
			dimensionSizes.push_back( size.asInt() );

		} // while

		scanner.match( ')' );

		// *******************************
		// Check for variable size column.
		// *******************************
		if( dimensionSizes.size() == 1 && dimensionSizes[ 0 ] == 0 )
		{
			dimensionSizes = Column::zero;
		} // if
	}
	else
	{
		dimensionSizes = Dal::scalar;
	}// if

	expression = ParseExpression( scanner );
	name = perlInterpreter.EvaluateExpression( expression, LookupTable );

	type = scanner.MatchOneOf(
			"int8",
			"uint16",
			"int16",
			"uint32",
			"int32",
			"real32",
			"real64",
			"bool",
			"string",
			(const char *)0 );

	scanner.Match( Token::String, label );
	scanner.Match( Token::String, units );

	column = table -> addColumn(
				name.asString(),
				toDalDataType( type.asString() ),
				label.asString(),
				units.asString(),
				dimensionSizes );

	// ********************************************************
	// Assign the new column to the lookup table variable "col"
	// "col" is really a pointer, but treat it as an integer
	// which should be ok (for this prototype version), but
	// later this should be implemented using a Pointer token.
	// ********************************************************
	ostrstream exp;
	exp << "col=pointer(" << int( column ) << ")" << ends;
	eval( exp.str(), &LookupTable );

	// ********************************************************
	// Check if column is to be filled with data.
	// This is achieved by checking to see if there are anymore
	// tokens in the line (EndOfFile() indicates this).
	// Then, if there are more tokens, then this must be the file name
	// and column number (which may be an expression, but must
	// evaluate to an integer); thus get the filename and column
	// number expression. Evaluate the expression, and then finally
	// fill the column (achieved with a call to fillColumn())
	// using the specified filename and column number.
	// ********************************************************
	if( !scanner.EndOfFile() )
	{
		expression = ParseExpression( scanner );
		expressionDetected = true;
	} // if

	scanner.SetState( ScannerT::NORMAL );

	if( scanner.NextIs( '<' ))
	{
		scanner.match( '<' );

		parseColumnDescriptorBody( scanner, column, postExpressions );

		scanner.match( '>' );
	} // if

	// ***********************************
	// Evaluate post expressions (if any).
	// ***********************************
	for( unsigned int i = 0; i < postExpressions.size(); ++i )
	{
		Token::BTree		expression;
		expression = postExpressions[ i ];
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // for

	if( expressionDetected )
	{
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // if

	ENDPROC

}


static void
parseTableDescriptorBody(
		ScannerT &	scanner,
		Table *		table,
		vector<Token::BTree> &	postExpressions )
{
	PROC

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.NextIs( "column" ))
		{
			parseColumnDescriptor( scanner, table );
		}
		else if( scanner.nextIs( perlBlock ))
		{
			parseExecutePerl( scanner );
		}
		else if( scanner.nextIs( dalBlock ))
		{
			parseExecuteDal( scanner );
		}
		else if( scanner.nextIs( postBlock ))
		{
			parsePostBlock( scanner, postExpressions );
		}
		else if( scanner.NextIs( "attribute" ))
		{
			parseAttributeDescriptor(
					scanner,
					table );
		}
		else if( scanner.NextIs( "comment" ))
		{
			parseCommentDescriptor(
					scanner,
					table );
		}
		else if( scanner.NextIs( "history" ))
		{
			parseHistoryDescriptor(
					scanner,
					table );
		}
		else if( scanner.NextIs( "include" ))
		{
			Token	fileName;

			// ************************************************************
			// The include directive must appear on a single line.
			// Get the line, and parse it.
			// ************************************************************
			scanner.SetState( ScannerT::LINE );
			scanner.match( "include" );
			scanner.Match( Token::String, fileName );
			scanner.SetState( ScannerT::NORMAL );

			// **********************************************************
			// Initialise a scanner and parse and execute the specified
			// include file.
			// **********************************************************
			ScannerT	s( fileName.asString() );
			parseTableDescriptorBody( s, table, postExpressions );
		}
		else
		{
			expectedOneOf(
				scanner,
				"column",
				"attribute",
				"comment",
				"history",
				"include",
				perlBlock.c_str(),
				dalBlock.c_str(),
				postBlock.c_str(),
				0 );
		} // if
	} // while

	ENDPROC
}


static void
parseTableDescriptor(
		ScannerT &	scanner,
		DataSet *	dataSet )
{

	PROC

	Token			token;
	Token			name;
	Token			label;
	Token			rows;
	Token::BTree		rows_expression;
	Token::BTree		expression;
	bool			expressionDetected = false;
	vector<Token::BTree>	postExpressions;

	// ************************************************************
	// The table descriptor must appear on a single line.
	// Get the line, and parse it.
	// ************************************************************
	scanner.SetState( ScannerT::LINE );
	scanner.match( "table" );
	expression = ParseExpression( scanner );
	name = perlInterpreter.EvaluateExpression( expression, LookupTable );

	rows_expression = ParseExpression( scanner );
	rows = perlInterpreter.EvaluateExpression( rows_expression, LookupTable );

	scanner.Match( Token::String, label );

	Table * newTable = dataSet -> addTable(
					name.asString(),
					rows.asInt(),
					label.asString() );

	// ********************************************************
	// Assign the new table to the lookup table variable "tab"
	// "tab" is really a pointer, but treat it as an integer
	// which should be ok (for this prototype version), but
	// later this should be implemented using a Pointer token.
	// ********************************************************
	ostrstream exp;
	exp << "tab=pointer(" << int( newTable ) << ")" << ends;
	eval( exp.str(), &LookupTable );


	// ***********************************
	// Check for the optional expression .
	// ***********************************
	if( !scanner.EndOfFile() )
	{
		expression = ParseExpression( scanner );
		expressionDetected = true;
	} // if

	scanner.SetState( ScannerT::NORMAL );

	if( scanner.NextIs( '<' ))
	{
		scanner.match( '<' );

		parseTableDescriptorBody( scanner, newTable, postExpressions );

		scanner.match( '>' );
	} // if

	// ***********************************
	// Evaluate post expressions (if any).
	// ***********************************
	for( unsigned int i = 0; i < postExpressions.size(); ++i )
	{
		Token::BTree		expression;
		expression = postExpressions[ i ];
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // for

	if( expressionDetected )
	{
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // if

	ENDPROC

}


static void
parseArrayDescriptorBody(
		ScannerT &	scanner,
		Array *		array,
		vector<Token::BTree> &	postExpressions )
{
	PROC

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.NextIs( "attribute" ))
		{
			parseAttributeDescriptor(
					scanner,
					array );
		}
		else if( scanner.nextIs( perlBlock ))
		{
			parseExecutePerl( scanner );
		}
		else if( scanner.nextIs( dalBlock ))
		{
			parseExecuteDal( scanner );
		}
		else if( scanner.nextIs( postBlock ))
		{
			parsePostBlock( scanner, postExpressions );
		}
		else if( scanner.NextIs( "comment" ))
		{
			parseCommentDescriptor(
					scanner,
					array );
		}
		else if( scanner.NextIs( "history" ))
		{
			parseHistoryDescriptor(
					scanner,
					array );
		}
		else if( scanner.NextIs( "include" ))
		{
			Token	fileName;

			// ************************************************************
			// The include directive must appear on a single line.
			// Get the line, and parse it.
			// ************************************************************
			scanner.SetState( ScannerT::LINE );
			scanner.match( "include" );
			scanner.Match( Token::String, fileName );
			scanner.SetState( ScannerT::NORMAL );

			// **********************************************************
			// Initialise a scanner and parse and execute the specified
			// include file.
			// **********************************************************
			ScannerT	s( fileName.asString() );
			parseArrayDescriptorBody( s, array, postExpressions );
		}
		else
		{
			expectedOneOf(
				scanner,
				"attribute",
				"comment",
				"history",
				"include",
				perlBlock.c_str(),
				dalBlock.c_str(),
				postBlock.c_str(),
				0 );
		} // if
	} // while
	ENDPROC
}


static void
parseArrayDescriptor(
		ScannerT &	scanner,
		DataSet *	dataSet )
{
	PROC

	Token			token;
	Token			name;
	Token			type;
	Token			size;
	Token			label;
	Token			units;
	Token::BTree		size_expression;
	vector<unsigned long>	dimensionSizes;
	Token::BTree		expression;
	bool			expressionDeteced = false;
	vector<Token::BTree>	postExpressions;

	// ************************************************************
	// The array descriptor must appear on a single line.
	// Get the line, and parse it.
	// ************************************************************
	scanner.SetState( ScannerT::LINE );
	scanner.match( "array" );

	scanner.match( '(' );
	size_expression = ParseAssignmentExpression( scanner );
	size = perlInterpreter.EvaluateExpression( size_expression, LookupTable );
	dimensionSizes.push_back( size.asInt() );

	while( scanner.NextIs( ',' ))
	{
		scanner.match( ',' );
		size_expression = ParseAssignmentExpression( scanner );
		size = perlInterpreter.EvaluateExpression( size_expression, LookupTable );
		dimensionSizes.push_back( size.asInt() );

	} // while

	scanner.match( ')' );

	expression = ParseExpression( scanner );
	name = perlInterpreter.EvaluateExpression( expression, LookupTable );

	type = scanner.MatchOneOf(
			"int8",
			"uint16",
			"int16",
			"uint32",
			"int32",
			"real32",
			"real64",
			(const char *)0 );

	scanner.Match( Token::String, label );
	scanner.Match( Token::String, units );

	Array * newArray = dataSet -> addArray(
				name.asString(),
				toDalDataType( type.asString() ),
				dimensionSizes,
				label.asString(),
				units.asString() );

	// ********************************************************
	// Assign the new array to the lookup table variable "arr"
	// "arr" is really a pointer, but treat it as an integer
	// which should be ok (for this prototype version), but
	// later this should be implemented using a Pointer token.
	// ********************************************************
	ostrstream exp;
	exp << "arr=pointer(" << int( newArray ) << ")" << ends;
	eval( exp.str(), &LookupTable );

	// **********************************
	// Check for the optional epxression.
	// **********************************
	if( !scanner.EndOfFile() )
	{
		expression = ParseExpression( scanner );
		expressionDeteced = true;
	} // if

	scanner.SetState( ScannerT::NORMAL );

	if( scanner.NextIs( '<' ))
	{
		scanner.match( '<' );

		parseArrayDescriptorBody( scanner, newArray, postExpressions );

		scanner.match( '>' );
	} // if

	// ***********************************
	// Evaluate post expressions (if any).
	// ***********************************
	for( unsigned int i = 0; i < postExpressions.size(); ++i )
	{
		Token::BTree		expression;
		expression = postExpressions[ i ];
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // for

	if( expressionDeteced )
	{
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // if

	ENDPROC

}


static void
parseDataSetDescriptorBody(
		ScannerT &		scanner,
		DataSet *		dataSet,
		vector<Token::BTree> &	postExpressions )
{
	PROC

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.NextIs( "table" ))
		{
			parseTableDescriptor( scanner, dataSet );
		}
		else if( scanner.nextIs( perlBlock ))
		{
			parseExecutePerl( scanner );
		}
		else if( scanner.nextIs( dalBlock ))
		{
			parseExecuteDal( scanner );
		}
		else if( scanner.nextIs( postBlock ))
		{
			parsePostBlock( scanner, postExpressions );
		}
		else if( scanner.NextIs( "array" ))
		{
			parseArrayDescriptor( scanner, dataSet );
		}
		else if( scanner.NextIs( "attribute" ))
		{
			parseAttributeDescriptor(
					scanner,
					dataSet );
		}
		else if( scanner.NextIs( "comment" ))
		{
			parseCommentDescriptor(
					scanner,
					dataSet );
		}
		else if( scanner.NextIs( "history" ))
		{
			parseHistoryDescriptor(
					scanner,
					dataSet );
		}
		else if( scanner.NextIs( "include" ))
		{
			Token	fileName;

			// ************************************************************
			// The include directive must appear on a single line.
			// Get the line, and parse it.
			// ************************************************************
			scanner.SetState( ScannerT::LINE );
			scanner.match( "include" );
			scanner.Match( Token::String, fileName );
			scanner.SetState( ScannerT::NORMAL );

			// **********************************************************
			// Initialise a scanner and parse and execute the specified
			// include file.
			// **********************************************************
			ScannerT	s( fileName.asString() );
			parseDataSetDescriptorBody( s, dataSet, postExpressions );
		}
		else
		{
			expectedOneOf(
				scanner,
				"table",
				"array",
				"attribute",
				"comment",
				"history",
				"include",
				perlBlock.c_str(),
				dalBlock.c_str(),
				postBlock.c_str(),
				0 );
		} // if
	} // while

	ENDPROC

}


static void
parseDataSetDescriptor( ScannerT & scanner )
{
	PROC

	Token			token;
	Token			name;
	Token::BTree		expression;
	bool			expressionDetected = false;
	vector<Token::BTree>	postExpressions;

	// ************************************************************
	// The data set descriptor must appear on a single line.
	// Get the line, and parse it.
	// ************************************************************
	scanner.SetState( ScannerT::LINE );
	scanner.match( "dataset" );
	expression = ParseExpression( scanner );
	name = perlInterpreter.EvaluateExpression( expression, LookupTable );

	DataSet	* newDataSet = dataSetServer -> open(
					name.asString(),
					DataSet::Create );

	// **********************************************************
	// Assign the new data set to the lookup table variable "dat"
	// "dat" is really a pointer, but treat it as an integer
	// which should be ok (for this prototype version), but
	// later this should be implemented using a Pointer token.
	// **********************************************************
	ostrstream exp;
	exp << "dat=pointer(" << int( newDataSet ) << ")" << ends;
	eval( exp.str(), &LookupTable );

	// **********************************
	// Check for the optional expression.
	// **********************************
	if( !scanner.EndOfFile() )
	{
		expression = ParseExpression( scanner );
		expressionDetected = true;
	} // if

	scanner.SetState( ScannerT::NORMAL );

	if( scanner.NextIs( '<' ))
	{
		scanner.match( '<' );

		parseDataSetDescriptorBody( scanner, newDataSet, postExpressions );

		scanner.match( '>' );
	} // if


	// ***********************************
	// Evaluate post expressions (if any).
	// ***********************************
	for( unsigned int i = 0; i < postExpressions.size(); ++i )
	{
		Token::BTree		expression;
		expression = postExpressions[ i ];
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // for

	if( expressionDetected )
	{
		perlInterpreter.EvaluateExpression( expression, LookupTable );
	} // if

	// ***************************
	// Finally close the data set.
	// ***************************
	dataSetServer -> close( newDataSet );

	ENDPROC;

}



static void
parseDescriptorFile( ScannerT & scanner )
{
	PROC

	while( !scanner.EndOfFile() )
	{
		if( scanner.NextIs( "dataset" ))
		{
			parseDataSetDescriptor( scanner );
		}
		else if( scanner.nextIs( perlBlock ))
		{
			parseExecutePerl( scanner );
		}
		else if( scanner.nextIs( dalBlock ))
		{
			parseExecuteDal( scanner );
		}
		else if( scanner.nextIs( "include" ))
		{
			Token	fileName;

			// ************************************************************
			// The include directive must appear on a single line.
			// Get the line, and parse it.
			// ************************************************************
			scanner.SetState( ScannerT::LINE );
			scanner.match( "include" );
			scanner.Match( Token::String, fileName );
			scanner.SetState( ScannerT::NORMAL );

			// **********************************************************
			// Initialise a scanner and parse and execute the specified
			// include file.
			// **********************************************************
			ScannerT	s( fileName.asString() );
			parseDescriptorFile( s );
		}
		else
		{
			expectedOneOf(
				scanner,
				"dataSet",
				"include",
				perlBlock.c_str(),
				dalBlock.c_str(),
				0 );
		} // if
	} // while

	ENDPROC

}

void deceit()
{
  PROC
  if (booleanParameter("withperllibrary"))
    perlInterpreter.initialise(stringParameter("perllibrary"));  
  else
    perlInterpreter.initialise(fileLocator.locate("lib/perl5/deceit.pl"));
  ScannerT scanner( stringParameter("file"));
  parseDescriptorFile( scanner );
  ENDPROC
}
